/*###################################################################################
#
#   Embperl - Copyright (c) 1997-2000 Gerald Richter / ECOS
#
#   You may distribute under the terms of either the GNU General Public
#   License or the Artistic License, as specified in the Perl README file.
#
#   THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
#   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#   WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
###################################################################################*/



struct tArrayCtrl
    {
    int	    nFill ;	    /* index of last element */
    int	    nMax ;	    /* number of last element for which is space */
    int	    nAdd ;	    /* number of elements to add on grow */
    int	    nElementSize ;  /* number of bytes per element */
    } ;

typedef void  tArray  ;

typedef unsigned char	tUInt8 ;
typedef signed char	tSInt8 ;
typedef unsigned short	tUInt16 ;
typedef signed short	tSInt16 ;
typedef unsigned long	tUInt32 ;
typedef signed long	tSInt32 ;

typedef tSInt32		tIndex ;
typedef tSInt16		tIndexShort ;
typedef tIndex		tStringIndex ;

typedef tUInt8          tNodeType ;
typedef tIndex		tNode	 ;
typedef tIndex		tAttr	 ;


struct tNodeData
    {
    tNodeType		nType ;
    tUInt8		bFlags ;
    tIndexShort		xDomTree ;
    tIndex		xNdx ;
    tStringIndex        nText ;
    tNode		xChilds ;
    tUInt16		numAttr ;
    tUInt16             nLinenumber ;
    tNode		xPrev ;
    tNode		xNext ;
    tNode		xParent ;
    } ;

typedef struct tNodeData tNodeData ;

struct tAttrData
    {
    tNodeType		nType ;     /* must be at the same offset as tNodeData.nType ! */
    tUInt8		bFlags ;
    tUInt16		nNodeOffset ;
    tIndex		xNdx ;      /* must be at the same offset as tNodeData.xNdx ! */
    tIndex              xName ;
    tIndex              xValue ;  /* must be at the same offset as tNodeData.xChilds ! */
    } ;

typedef struct tAttrData tAttrData ;


struct tDomNode
    {
    tIndex		xDomTree ;
    tNode		xNode ;
    SV *		pDomNodeSV ;
    } ;

typedef struct tDomNode tDomNode ;


/*
  Node Types
*/

/* ------------------------------------------------------------------------ 

interface Node {
  // NodeType
  const unsigned short      ELEMENT_NODE                   = 1;
  const unsigned short      ATTRIBUTE_NODE                 = 2;
  const unsigned short      TEXT_NODE                      = 3;
  const unsigned short      CDATA_SECTION_NODE             = 4;
  const unsigned short      ENTITY_REFERENCE_NODE          = 5;
  const unsigned short      ENTITY_NODE                    = 6;
  const unsigned short      PROCESSING_INSTRUCTION_NODE    = 7;
  const unsigned short      COMMENT_NODE                   = 8;
  const unsigned short      DOCUMENT_NODE                  = 9;
  const unsigned short      DOCUMENT_TYPE_NODE             = 10;
  const unsigned short      DOCUMENT_FRAGMENT_NODE         = 11;
  const unsigned short      NOTATION_NODE                  = 12;

*/

enum tNodeType
    {
    ntypTag	        = 1,
    ntypStartTag        = 1 + 0x20,
    ntypEndTag	        = 1 + 0x40,
    ntypEndStartTag     = 1 + 0x60,
    ntypAttr	        = 2,
    ntypAttrValue       = 2 + 0x20,
    ntypText	        = 3,
    ntypCDATA	        = 4,
    ntypEntityRef       = 5,
    ntypEntity          = 6,
    ntypProcessingInstr = 7,
    ntypComment         = 8,
    ntypDocument        = 9,
    ntypDocumentType    = 10,
    ntypDocumentFraq    = 11,
    ntypNotation        = 12,
    } ;

#define ntypMask 0x1f 

enum tNodeFlags
    {
    nflgDeleted	    = 0,
    nflgOK	    = 1,
    nflgEscUrl      = 2,
    nflgEscChar	    = 4,
    nflgIgnore      = 8,
    nflgModified    = 16,
    nflgCheckpoint  = 32,
    nflgReturn      = 64,
    } ;

enum tAttrFlags
    {
    aflgDeleted	    = 0,
    aflgOK	    = 1,
    aflgAttrValue   = 2,
    aflgAttrChilds  = 4,
    } ;

struct tDomTreeOrder
    {
    tNode  xFromNode ;	    
    tNode  xToNode ;
    } ;

typedef struct tDomTreeOrder tDomTreeOrder ;


struct tDomTree
    {
    void * *	    pLookup ;	/* table for converting tNode and tAttr to pointers */
    tDomTreeOrder * pOrder ;	/* Order of dom tree after execution of code */
    tIndex	    xNdx ;	/* Index of Dom Tree */
    tNode	    xDocument ;	/* root document node */
    tNode           xLastNode ; /* last node that was compiled */
    tNode           xCurrNode ; /* curr node that is compiled */
    tIndex          xFilename ; /* name of source file */
    SV *	    pSV ;	/* general purpose SV */
    SV *	    pDomTreeSV ; /* SV that's hold the Index */
				 /* Domtree will be deleted when this SV is delted */
    } ;

typedef struct tDomTree tDomTree ;




extern tDomTree  *    pDomTrees ;	     /* Array with all Dom Trees */
extern HE * *	      pStringTableArray  ;   /* Array with pointers to strings */
extern tIndex	      xNoName ;		     /* String index for Attribut with noname */
extern tIndex	      xDomTreeAttr ;	     /* String index for Attribut which holds the DomTree index */
extern tIndex	      xDocument ;	     /* String index for Document */
extern tIndex	      xDocumentFraq ;	     /* String index for Document Fraquent */
extern tIndex	      xOrderIndexAttr ;	     /* String index for Attribute which holds the OrderIndex */

extern int nCheckpointCache ;
extern int nCheckpointCacheMask ;
extern tIndex xCheckpointCache[] ;


tStringIndex String2NdxInc (/*in*/ const char *	    sText,
			    /*in*/ int		    nLen,
			    /*in*/ int		    bInc) ;


#define SV2String(pSV,l)  (SvOK (pSV)?SvPV (pSV, l):(l=0,NULL))

#define String2NdxNoInc(sText,nLen) String2NdxInc(sText,nLen,0)
#define String2Ndx(sText,nLen) String2NdxInc(sText,nLen,1)

#define Ndx2String(nNdx) (HeKEY (pStringTableArray[nNdx]))
#define Ndx2StringLen(nNdx,sVal,nLen) { HE * pHE = pStringTableArray[nNdx] ; nLen=HeKLEN(pHE) ; sVal = HeKEY (pHE) ; }

#define NdxStringRefcntInc(nNdx) (SvREFCNT_inc (HeVAL (pStringTableArray[nNdx])))


int ArrayNew (/*in*/ const tArray * pArray,
	      /*in*/ int	nAdd,
	      /*in*/ int	nElementSize) ;

int ArrayFree (/*in*/ const tArray * pArray) ;

int ArraySet (/*in*/ const tArray * pArray,
	      /*in*/ int	numElements) ;

int ArraySetSize (/*in*/ const tArray * pArray,
	          /*in*/ int	numElements) ;

int ArrayGetSize (/*in*/ const tArray * pArray) ;

void StringNew (/*in*/ char * * pArray,
	      /*in*/ int	nAdd) ;

void StringFree (/*in*/ char * * pArray) ;

int StringAdd (/*in*/ char * *	   pArray,
	       /*in*/ const char * sAdd,
	       /*in*/ int 	   nLen) ;

int DomInit (void) ;

void DomStats (void) ;

int DomTree_clone (/*in*/ tDomTree *	pOrgDomTree,
		   /*out*/tDomTree * *  pNewDomTree,
		   /*in*/ int           bForceDocFraq) ;

int DomTree_new (tDomTree * * pNewLookup) ;

int DomTree_delete (tDomTree * pDomTree) ;

void DomTree_checkpoint (tIndex xDomTree, tNode xChild) ;
void DomTree_selfCheckpoint (tDomTree * pDomTree, tNode xFromNode, tNode xToNode) ;

int DomTree_selfDiscardAfterCheckpoint (/*in*/ tDomTree *	    pDomTree, 
				         /*in*/ tNodeData *     pNode,
				         /*in*/ tNodeData *     pArgNode) ;

int DomTree_discardAfterCheckpoint (/*in*/ tIndex  xDomTree, 
				     /*in*/ tNode   xNode) ;

#define DomTree_SV(pSV) SvIVX(pSV)
#define DomTree_selfSV(pSV) (DomTree_self(SvIVX(pSV)))
#define SV_DomTree(xDomTree) (DomTree_self (xDomTree) -> pDomTreeSV)
#define SV_DomTree_self(pDomTree) (pDomTree -> pDomTreeSV)

tNode Node_appendChild (/*in*/ tDomTree *  pDomTree,
			/*in*/	tNodeType	 nType,
			/*in*/	int              bForceAttrValue,
			/*in*/	const char *	 sText,
			/*in*/	int		 nTextLen,
			/*in*/	tNode 		 xParent,
			/*in*/	int		 nLevel,
			/*in*/	int		 nLinenumber) ;



tNodeData * Node_selfNthChild (/*in*/ tDomTree *  pDomTree,
				      /*in*/ struct tNodeData * pNode,
				      /*in*/ int         nChildNo) ;

tNodeData * Node_selfLastChild   (/*in*/  tDomTree *   pDomTree,
				  /*in*/  tNodeData *  pNode) ;

tNodeData * Node_selfNextSibling (/*in*/ tDomTree *  pDomTree,
			          /*in*/ tNodeData * pNode) ;
				  
tNode Node_nextSibling (/*in*/ tDomTree *  pDomTree,
			/*in*/ tNode       xNode) ;

#define DomTree_self(xDomTree)		    (&pDomTrees[xDomTree]) 

#define Node_self(pDomTree,xNode)	    ((struct tNodeData *)(pDomTree -> pLookup[xNode]))

#define Node_parentNode(pDomTree,xNode)	    (Node_self(pDomTree,xNode)->xParent)
#define Node_selfParentNode(pDomTree,pNode) (Node_self(pDomTree,(pNode)->xParent))
#define xNode_selfParentNode(pDomTree,pNode) ((pNode)->xParent)

#define Node_firstChild(pDomTree,xNode)	    (Node_self(pDomTree,xNode)->xChilds)
#define Node_selfFirstChild(pDomTree,pNode)  (Node_self(pDomTree,(pNode)->xChilds))

#define Node_selfNodeNameNdx(pNode)	    ((pNode) -> nText) ;
#define Node_selfNodeName(pNode)	    (Ndx2String ((pNode) -> nText))
#define Node_nodeName(pDomTree,pNode)	    (Ndx2String (Node_self (pDomTree,xNode) -> nText))
#define Node_selfFirstAttr(pNode)	    ((tAttrData *)((pNode) + 1))

tNodeData * Node_selfCloneNode (/*in*/ tDomTree *      pDomTree,
				/*in*/ tNodeData *     pNode,
				/*in*/  int	       bDeep) ;

tNode Node_cloneNode (/*in*/ tDomTree *      pDomTree,
		    /*in*/ tNode 	    xNode,
		    /*in*/  int		    bDeep) ;


void Node_toString (/*in*/ tDomTree *  pDomTree,
		    /*i/o*/ register req * r,
		    /*in*/ tNode	    xNode) ;

struct tNodeData *  Node_selfRemoveChild (/*in*/ tDomTree *  pDomTree,
					  /*in*/ tNode		    xNode,
					  /*in*/ struct tNodeData *	    pChild) ;

tNode Node_removeChild (/*in*/ tDomTree *  pDomTree,
			/*in*/ tNode	    xNode,
			/*in*/ tNode	    xChild) ;

tNode Node_insertAfter          (/*in*/ tDomTree *      pNewNodeDomTree,
				 /*in*/ tNode		xNewNode,
                                 /*in*/ tDomTree *      pRefNodeDomTree,
				 /*in*/ tNode		xRefNode) ;

tNode Node_replaceChildWithNode (/*in*/ tDomTree *      pParentDomTree,
			         /*in*/	tNode 	        xParent,
                                 /*in*/ tDomTree *      pChildDomTree,
			         /*in*/	tNode 	        xChild) ;

tNode Node_replaceChildWithCDATA (/*in*/ tDomTree *  pDomTree,
				  /*in*/ tNode		 xNode,
				  /*in*/ tNode		 xOldChild,
			          /*in*/ const char *	 sText,
				  /*in*/ int		 nTextLen,
				  /*in*/ int		 nEscMode,
				  /*in*/ int		 bFlags) ;




tAttrData *  Element_selfGetAttribut (/*in*/ tDomTree *     pDomTree,
				      /*in*/ struct tNodeData * pNode,
				      /*in*/ const char *	sAttrName,
				      /*in*/ int		nAttrNameLen) ;



tAttrData *  Element_selfGetNthAttribut (/*in*/ tDomTree *	   pDomTree,
			  	         /*in*/ struct tNodeData * pNode,
				         /*in*/ int                n) ;


tAttrData *  Element_selfSetAttribut (/*in*/ tDomTree *	        pDomTree,
				      /*in*/ struct tNodeData * pNode,
				      /*in*/ const char *	sAttrName,
				      /*in*/ int		nAttrNameLen,
				      /*in*/ const char *       sNewValue, 
				      /*in*/ int		nNewValueLen,
				      /*in*/ int		bClone) ;


tAttrData *  Element_selfRemoveAttribut (/*in*/ tDomTree *	        pDomTree,
				      /*in*/ struct tNodeData * pNode,
				      /*in*/ const char *	sAttrName,
				      /*in*/ int		nAttrNameLen,
				      /*in*/ int		bClone) ;

#define Attr_self(pDomTree,xAttr)	    ((struct tAttrData *)(pDomTree -> pLookup[xAttr]))
#define Attr_selfNode(pAttr)        ((struct tNodeData * )(((tUInt8 *)pAttr) - pAttr -> nNodeOffset))
#define Attr_selfAttrNum(pAttr)    (pAttr - Node_selfFirstAttr (Attr_selfNode(pAttr)))


char *       Attr_selfValue (/*in*/ tDomTree *	        pDomTree,
			     /*in*/ struct tAttrData * pAttr,
			    /*out*/ char * *    ppAttr) ;
